"""
association measures

"""

from warnings import warn

import numpy as np
from pandas import concat, merge
from scipy.stats import beta, norm

from .binomial import choose
from .frequencies import expected_frequencies, observed_frequencies

CHOOSE = np.vectorize(choose)


def list_measures():
    """Return a dictionary of implemented measures (name: measure)

    :return: dictionary of measures
    :rtype: dict
    """

    return {
        # asymptotic hypothesis tests
        'z_score': z_score,
        't_score': t_score,
        'log_likelihood': log_likelihood,
        'simple_ll': simple_ll,
        # point estimates of association strength
        'min_sensitivity': min_sensitivity,
        'liddell': liddell,
        'dice': dice,
        'log_ratio': log_ratio,
        # likelihood measures
        # 'hypergeometric_likelihood': hypergeometric_likelihood,
        # 'binomial_likelihood': binomial_likelihood,
        # conservative estimates
        'conservative_log_ratio': conservative_log_ratio,
        # information theory
        'mutual_information': mutual_information,
        'local_mutual_information': local_mutual_information,
    }


def score(df, measures=None, f1=None, N=None, N1=None, N2=None,
          freq=True, per_million=True, digits=6, disc=.001,
          discounting='Walter1975', signed=True, alpha=.001,
          correct='Bonferroni', boundary='poisson', vocab=None,
          one_sided=False):
    """Calculate a list of association measures on columns of df. Defaults
    to all available (and numerically stable) measures.

    Columns of df must follow one of the following notations:
    - contingency table: O11, O12, O21, O22
    - frequency signature: f, f1, f2, N
    - corpus frequencies: f1, N1, f2, N2
    Integers (f1, N, N1, N2) can also be passed as scalar arguments. See
    frequencies.observed_frequencies for further info on notation.

    :param DataFrame df: Dataframe with reasonably-named frequency columns
    :param list measures: names of measures (or measures)
    :param bool freq: also return observed and expected frequencies (incl. marginals)?
    :param bool per_million: return instances per million? (only if freq is True)
    :param int digits: round scores

    Further keyword arguments will be passed to the respective measures:
    :param float disc: discounting (or smoothing) parameter for O11 == 0 (and O21 == 0)
    :param str discounting: LR: discounting strategy (Walter1975 vs. Hardie2014)
    :param bool signed: enforce negative values for rows with O11 < E11?
    :param float alpha: CLR: significance level
    :param str boundary: CLR: exact CI boundary of [poisson] distribution or [normal] approximation?
    :param str correct: CLR: correction type repeated tests (None|"Bonferroni"|"Sidak")
    :param int vocab: CLR: size of vocabulary (number of comparisons for correcting alpha)
    :param bool one_sided: CLR: calculate one- or two-sided confidence interval

    :return: association measures
    :rtype: DataFrame

    """

    # convert input to contingency notation and calculate expected frequencies
    df = observed_frequencies(df, f1=f1, N=N, N1=N1, N2=N2)
    df = expected_frequencies(df, observed=True)
    freq_columns = df.columns

    # select measures
    ams_all = list_measures()
    if measures is not None:
        if isinstance(measures[0], str):
            # TODO issue warning if measure not in list
            measures = [ams_all[k] for k in measures if k in ams_all.keys()]
    else:
        measures = [ams_all[k] for k in ams_all]

    # reduce df to unique frequency signatures
    vocab = len(df) if vocab is None else vocab
    df_reduced = df.drop_duplicates(subset=list(freq_columns)).copy()

    # calculate measures
    for measure in measures:
        df_reduced[measure.__name__] = measure(
            df_reduced, disc=disc, discounting=discounting, signed=signed, alpha=alpha,
            correct=correct, boundary=boundary, vocab=vocab, one_sided=one_sided
        )

    # join on frequency columns (NB: thanks to pandas API, we have to take care of index names ourselves)
    index_names = ['index'] if df.index.names == [None] else df.index.names
    df = merge(df.reset_index(), df_reduced, how='left', on=list(freq_columns)).set_index(index_names)

    # keep frequency columns?
    if not freq:
        df = df.drop(freq_columns, axis=1)
    else:
        # add instances (per million)
        fac = 10**6 if per_million else 1
        name = 'ipm' if per_million else 'instances'
        df[name] = df['O11'] / df['R1'] * fac
        df[name + '_reference'] = df['O21'] / df['R2'] * fac
        df[name + '_expected'] = df['E11'] / df['R1'] * fac

    # rounding
    df = round(df, digits) if digits is not None else df

    return df


def calculate_measures(df, measures=None, freq=False, per_million=True, digits=None, **kwargs):
    """deprecated since 0.2.3, use `score()` instead.

    """
    warn(
        "`calculate_measures()` is deprecated since version 0.2.3, use `score()` instead",
        DeprecationWarning,
        stacklevel=2
    )
    return score(df, measures=measures, freq=freq, per_million=per_million, digits=digits, **kwargs)


###############################
# ASYMPTOTIC HYPOTHESIS TESTS #
###############################

def z_score(df, **kwargs):
    """Calculate z-score

    :param DataFrame df: DataFrame with columns O11 and E11
    :return: z-score
    :rtype: pd.Series
    """

    am = (df['O11'] - df['E11']) / np.sqrt(df['E11'])

    return am


def t_score(df, disc=.001, **kwargs):
    """Calculate t-score

    :param DataFrame df: pd.DataFrame with columns O11 and E11
    :param float disc: discounting (or smoothing) parameter for O11 == 0
    :return: t-score
    :rtype: pd.Series
    """

    O11_disc = df['O11'].where(df['O11'] != 0, disc)
    am = (df['O11'] - df['E11']) / np.sqrt(O11_disc)

    return am


def log_likelihood(df, signed=True, **kwargs):
    """Calculate log-likelihood

    :param DataFrame df: pd.DataFrame with columns O11..O22, E11..E22
    :param bool signed: return negative values for rows with O11 < E11?
    :return: log-likelihood
    :rtype: pd.Series
    """

    # NB: discounting will not have any effect:
    #     term will be multiplied by original Oij = 0
    O11_disc = df['O11'].where(df['O11'] != 0, 1)
    O12_disc = df['O12'].where(df['O12'] != 0, 1)
    O21_disc = df['O21'].where(df['O21'] != 0, 1)
    O22_disc = df['O22'].where(df['O22'] != 0, 1)

    ii = df['O11'] * np.log(O11_disc / df['E11'])
    ij = df['O12'] * np.log(O12_disc / df['E12'])
    ji = df['O21'] * np.log(O21_disc / df['E21'])
    jj = df['O22'] * np.log(O22_disc / df['E22'])

    am = 2 * (ii + ij + ji + jj)

    if signed:
        am = np.sign(df['O11'] - df['E11']) * am

    return am


def simple_ll(df, signed=True, **kwargs):
    """Calculate simple log-likelihood

    :param DataFrame df: pd.DataFrame with columns O11, E11
    :param bool signed: return negative values for rows with O11 < E11?
    :return: simple log-likelihood
    :rtype: pd.Series
    """

    # NB: discounting will not have any effect: term will be multiplied by original Oij = 0
    O11_disc = df['O11'].where(df['O11'] != 0, 1)

    log_term = df['O11'] * np.log(O11_disc / df['E11'])

    am = 2 * (log_term - (df['O11'] - df['E11']))

    if signed:
        am = np.sign(df['O11'] - df['E11']) * am

    return am


###########################################
# POINT ESTIMATES OF ASSOCIATION STRENGTH #
###########################################

def min_sensitivity(df, **kwargs):
    """Calculate Minimum Sensitivity.

    :param DataFrame df: pd.DataFrame with columns O11, R1, C1
    :return: dice
    :rtype: pd.Series
    """

    am1 = df['O11'] / df['R1']
    am2 = df['O11'] / df['C1']
    am = concat([am1, am2], axis=1).min(axis=1)

    return am


def liddell(df, **kwargs):
    """Calculate Liddell

    :param DataFrame df: pd.DataFrame with columns O11, O12, O21, O22, C1, C2
    :return: liddell
    :rtype: pd.Series
    """

    am = (df['O11'] * df['O22'] - df['O12'] * df['O21']) / df['C1'] / df['C2']

    return am


def dice(df, **kwargs):
    """Calculate Dice coefficient

    :param DataFrame df: pd.DataFrame with columns O11, O12, O21
    :return: dice
    :rtype: pd.Series
    """

    am = (2 * df['O11']) / (2 * df['O11'] + df['O12'] + df['O21'])

    return am


def log_ratio(df, disc=.5, discounting='Walter1975', **kwargs):
    """Calculate log-ratio, i.e. binary logarithm of relative risk

    :param DataFrame df: pd.DataFrame with columns O11, O21, R1, R2
    :param float disc: discounting (or smoothing) parameter for O11 == 0 and O21 == 0
    :param str discounting: discounting according to Walter1975 or Hardie2014?
    :return: log-ratio
    :rtype: pd.Series
    """

    if discounting == 'Walter1975':
        # mathematically sensible discounting according to Walter (1975)
        am = np.log2(((df['O11'] + disc) / (df['R1'] + disc)) / ((df['O21'] + disc) / (df['R2'] + disc)))

    elif discounting == 'Hardie2014':
        # questionable discounting according to Hardie (2014)
        O11_disc = df['O11'].where(df['O11'] != 0, disc)
        O21_disc = df['O21'].where(df['O21'] != 0, disc)
        am = np.log2((O11_disc / O21_disc) / (df['R1'] / df['R2']))

    return am


#######################
# LIKELIHOOD MEASURES #
#######################

def hypergeometric_likelihood(df, **kwargs):
    """Calculate hypergeometric-likelihood

    :param DataFrame df: pd.DataFrame with columns O11, O12, O21, O22
    :return: hypergeometric-likelihood
    :rtype: pd.Series
    """

    df = df.astype(
        {'O11': 'int32', 'O12': 'int32', 'O21': 'int32', 'O22': 'int32'}
    )

    np.seterr(all='ignore')
    c1 = CHOOSE(df['O11'] + df['O21'], df['O11'])
    c2 = CHOOSE(df['O12'] + df['O22'], df['O12'])
    c3 = CHOOSE(df['O11'] + df['O12'] + df['O21'] + df['O22'], df['O11'] + df['O12'])
    am = c1 / c3 * c2
    np.seterr(all='warn')

    return am


def binomial_likelihood(df, **kwargs):
    """Calculate binomial-likelihood

    :param DataFrame df: pd.DataFrame with columns O11, O12, O21, O22, E11, N
    :return: binomial-likelihood
    :rtype: pd.Series
    """

    df = df.astype(
        {'O11': 'int32', 'O12': 'int32', 'O21': 'int32', 'O22': 'int32', 'E11': 'int32'}
    )

    np.seterr(all='ignore')
    c1 = CHOOSE(df['N'], df['O11'])
    c2 = (df['E11'] / df['N']) ** df['O11']
    c3 = (1 - df['E11'] / df['N']) ** (df['N'] - df['O11'])
    am = c1 * c2 * c3
    np.seterr(all='warn')

    return am


##########################
# CONSERVATIVE ESTIMATES #
##########################

def conservative_log_ratio(df, disc=.5, alpha=.001, boundary='poisson',
                           correct='Bonferroni', vocab=None,
                           one_sided=False, **kwargs):
    """Calculate conservative log-ratio, i.e. the binary logarithm of the
    lower bound of the confidence interval of relative risk at the
    (Bonferroni-corrected) significance level.

    :param DataFrame df: pd.DataFrame with columns O11, O12, O21, O22
    :param float disc: discounting (or smoothing) parameter for O11 == 0 and O21 == 0
    :param float alpha: significance level
    :param str boundary: exact CI boundary of [poisson] distribution or [normal] approximation?
    :param str correct: correction type for several tests (None | "Bonferroni" | "Sidak")
    :param int vocab: size of vocabulary (number of comparisons for correcting alpha)
    :param bool one_sided: calculate one- or two-sided confidence interval

    :return: conservative log-ratio
    :rtype: pd.Series
    """

    # correction of alpha for two-sided tests
    if not one_sided:
        alpha /= 2

    # Bonferroni or Sidak correction
    if correct is not None:
        if isinstance(correct, str):
            vocab = (df['O11'] >= 1).sum() if vocab is None else vocab
            if correct == 'Bonferroni':
                alpha /= vocab
            elif correct == "Sidak":
                alpha = 1 - (1 - alpha) ** (1 / vocab)
                # more stable alternative: alpha = 1 - exp(log(1 - alpha) / vocab)
                # doesn't make any difference in practice though, e.g. alpha = .00001, vocab = 10**10
            else:
                raise ValueError('parameter "correct" should either be "Bonferroni" or "Sidak".')
        else:
            raise ValueError('parameter "correct" should either be None or a string.')

    # CONFIDENCE INTERVAL

    # Poisson approximation (Evert 2022)
    if boundary == 'poisson':
        # only calculate where_lower
        lower = beta.ppf(alpha, df['O11'], df['O21'] + 1)
        lower_boundary = np.log2((df['R2'] / df['R1']) * lower / (1 - lower)).clip(lower=0)
        # only calculate where_upper
        upper = beta.ppf(1 - alpha, df['O11'] + 1, df['O21'])
        upper_boundary = np.log2((df['R2'] / df['R1']) * upper / (1 - upper)).clip(upper=0)
        # combine, set to 0 where (df['O11'] == 0) & (df['O12'] == 0)
        clrr = lower_boundary.where(
            (df['O11'] / df['R1']) >= (df['O21'] / df['R2']),
            upper_boundary
        )
        clrr = clrr.where(~((df['O11'] == 0) & (df['O12'] == 0)), 0).fillna(0)

    # Normal approximation (Hardie 2014)
    elif boundary == 'normal':
        # - questionable discounting according to Hardie (2014)
        O11_disc = df['O11'].where(df['O11'] != 0, disc)
        O21_disc = df['O21'].where(df['O21'] != 0, disc)
        # - compute natural logarithm of relative risk so we can use estimate for standard error of log(RR)
        lrr = np.log((O11_disc / O21_disc) / (df['R1'] / df['R2']))
        # - asymptotic standard deviation of log(RR) according to Wikipedia
        lrr_sd = np.sqrt(1/O11_disc + 1/O21_disc - 1/df['R1'] - 1/df['R2'])
        # - calculate and apply appropriate boundary
        z_factor = norm.ppf(1 - alpha)
        ci_min = (lrr - lrr_sd * z_factor).clip(lower=0)
        ci_max = (lrr + lrr_sd * z_factor).clip(upper=0)
        clrr = ci_min.where(lrr >= 0, ci_max)
        clrr /= np.log(2)           # adjust to binary logarithm

    return clrr


######################
# INFORMATION THEORY #
######################

def mutual_information(df, disc=.001, **kwargs):
    """Calculate Mutual Information

    :param DataFrame df: pd.DataFrame with columns O11 and E11
    :param float disc: discounting (or smoothing) parameter for O11 == 0
    :return: mutual information
    :rtype: pd.Series
    """

    O11_disc = df['O11'].where(df['O11'] != 0, disc)
    am = np.log10(O11_disc / df['E11'])

    return am


def local_mutual_information(df, **kwargs):
    """Calculate Local Mutual Information

    :param DataFrame df: pd.DataFrame with columns O11 and E11
    :return: local mutual information
    :rtype: pd.Series
    """

    # NB: discounting will not have any effect: term will be multiplied by original Oij = 0
    O11_disc = df['O11'].where(df['O11'] != 0, 1)
    am = df['O11'] * np.log10(O11_disc / df['E11'])

    return am



"""
comparison of ranked lists

"""


# from rbo import RankingSimilarity


# def compare_rbo(df1, df2, p):

#     return RankingSimilarity(df1.index, df2.index).rbo(p=p)


# def compare_gwets_ac1(df1, df2):

#     pass



#!/usr/bin/env python3
#cython: language_level=3
#cython: boundscheck=False
#cython: wraparound=False

cimport scipy.linalg
from scipy.linalg.cython_blas cimport cdotu
from scipy.linalg.cython_lapack cimport dgtsv

cpdef tridiag(double[:] a, double[:] b, double[:] c, double[:] x):
    """ Solve the system A y = x for y where A is the tridiagonal matrix with
    subdiagonal 'a', diagonal 'b', and superdiagonal 'c'. """
    cdef int n=b.shape[0], nrhs=1, info
    # Solution is written over the values in x.
    dgtsv(&n, &nrhs, &a[0], &b[0], &c[0], &x[0], &n, &info)

cpdef float complex complex_dot(float complex[:] cx, float complex[:] cy):
    """ Take dot product of two complex vectors """
    cdef:
        int n = cx.shape[0]
        int incx = cx.strides[0] // sizeof(cx[0])
        int incy = cy.strides[0] // sizeof(cy[0])
    return cdotu(&n, &cx[0], &incx, &cy[0], &incy)



import association_measures.frequencies as fq
import association_measures.measures as am
from association_measures.grids import topography


def test_frequencies(ucs_dataframe):

    # contingency table notation
    df = ucs_dataframe
    obs = fq.observed_frequencies(df)
    print()
    print(obs[['O11', 'O12', 'O21', 'O22']].head())

    # frequency signature notation
    df.rename({'l2': 'item'}, axis=1, inplace=True)
    df = df[['item', 'f', 'f1', 'f2', 'N']]
    df.index.name = 'id'
    print()
    print(df.head())

    # keywords
    tmp = df[['item', 'f', 'f1']].rename({'f': 'f1', 'f1': 'N1'}, axis=1)
    tmp['f2'] = df['f2'] - df['f']
    tmp['N2'] = df['N'] - df['f1']
    print()
    print(tmp.head())

    # expected frequencies
    exp = fq.expected_frequencies(df)
    print()
    print(exp[['E11', 'E12', 'E21', 'E22']].head())

    # observed frequencies
    obs = fq.observed_frequencies(df)
    print()
    print(obs[['O11', 'O12', 'O21', 'O22']].head())

    # raw
    print()
    print(df.head())


def test_measures(ucs_dataframe):

    df = fq.expected_frequencies(ucs_dataframe, observed=True)
    df['item'] = ucs_dataframe['l2']
    df = df.set_index('item')

    df_ams = am.log_likelihood(df)
    print(df_ams.head())

    df_ams = am.score(df, measures=['log_likelihood'])
    print(df_ams.head())

    df_ams = am.score(df, freq=False)
    print(df_ams.head())

    df_ams = am.score(df, measures=['log_likelihood'], signed=False, freq=False)
    print(df_ams.head())


def test_topography():

    print(topography())



import association_measures.measures as am
from pandas import DataFrame


def test_nas():

    d = DataFrame(data={'O11': 0, 'O12': 10, 'O21': 10, 'O22': 100}, index=['test'])
    scores = am.score(d)
    print(scores[['mutual_information', 'local_mutual_information']])

    d = DataFrame(data={'O11': 0, 'O12': 10, 'O21': 0, 'O22': 100}, index=['test'])
    scores = am.score(d)
    print(scores[['ipm', 'ipm_expected']])



import pytest
from numpy import isnan


import association_measures.measures as am
import association_measures.frequencies as fq


######
# MI #
######

@pytest.mark.mi
def test_mutual_information_single(fixed_dataframe):

    df = fq.expected_frequencies(fixed_dataframe, observed=True)
    df_ams = am.mutual_information(df)
    assert df_ams[0] == 1.0


@pytest.mark.mi
def test_mutual_information(fixed_dataframe):

    df = fixed_dataframe
    df_ams = am.score(df, ['mutual_information'])
    assert df_ams['mutual_information'][0] == 1.0


@pytest.mark.mi
@pytest.mark.invalid
def test_mutual_information_invalid(invalid_dataframe):

    df = invalid_dataframe
    with pytest.raises(ValueError):
        am.score(df, ['mutual_information'])


@pytest.mark.mi
@pytest.mark.zero
def test_mutual_information_zero(zero_dataframe):

    df = zero_dataframe
    df_ams = am.score(df, ['mutual_information'])
    assert df_ams['mutual_information'].iloc[0] == 0.061675


############
# LOCAL MI #
############

@pytest.mark.local_mi
def test_local_mi_single(fixed_dataframe):

    df = fq.expected_frequencies(fixed_dataframe, observed=True)
    df_ams = am.local_mutual_information(df)
    assert df_ams[0] == 10.0


@pytest.mark.local_mi
def test_local_mi(fixed_dataframe):

    df = fixed_dataframe
    df_ams = am.score(df, ['local_mutual_information'])
    assert df_ams['local_mutual_information'][0] == 10.0


@pytest.mark.local_mi
@pytest.mark.invalid
def test_local_mi_invalid(invalid_dataframe):

    df = invalid_dataframe
    with pytest.raises(ValueError):
        am.score(df, ['local_mutual_information'])


@pytest.mark.local_mi
@pytest.mark.zero
def test_local_mi_zero(zero_dataframe):

    df = zero_dataframe
    df_ams = am.score(df, ['local_mutual_information'], freq=True)
    assert df_ams['local_mutual_information'].iloc[0] == 848.954896


########
# DICE #
########

@pytest.mark.dice
def test_dice(fixed_dataframe):

    df = fixed_dataframe
    df_ams = am.score(df, ['dice'])

    assert df_ams['dice'][0] == 1.0


@pytest.mark.dice
@pytest.mark.invalid
def test_dice_invalid(invalid_dataframe):

    df = invalid_dataframe
    with pytest.raises(ValueError):
        am.score(df, ['dice'])


@pytest.mark.dice
@pytest.mark.zero
def test_dice_zero(zero_dataframe):
    df = zero_dataframe
    df_ams = am.score(df, ['dice'])
    assert df_ams['dice'].iloc[0] == 0.168312


##########
# TSCORE #
##########

@pytest.mark.t_score
def test_t_score_single(fixed_dataframe):

    df = fq.expected_frequencies(fixed_dataframe, observed=True)
    m = am.t_score(df)
    assert m[0] == 2.846049894151541


@pytest.mark.t_score
def test_t_score(fixed_dataframe):

    df = fixed_dataframe
    df_ams = am.score(df, ['t_score'])
    assert df_ams['t_score'][0] == 2.84605


@pytest.mark.t_score
@pytest.mark.invalid
def test_t_score_invalid(invalid_dataframe):
    df = invalid_dataframe
    with pytest.raises(ValueError):
        am.score(df, ['t_score'])


@pytest.mark.t_score
@pytest.mark.zero
def test_t_score_zero(zero_dataframe):
    df = zero_dataframe
    df_ams = am.score(df, ['t_score'], disc=.5)
    assert df_ams['t_score'].iloc[0] == 15.532438


##########
# ZSCORE #
##########

@pytest.mark.z_score
def test_z_score(fixed_dataframe):

    df = fixed_dataframe
    df_ams = am.score(df, ['z_score'])
    assert df_ams['z_score'][0] == 9.0


@pytest.mark.z_score
@pytest.mark.invalid
def test_z_score_nan(invalid_dataframe):
    df = invalid_dataframe
    with pytest.raises(ValueError):
        am.score(df, ['z_score'])


@pytest.mark.z_score
@pytest.mark.zero
def test_z_score_zero(zero_dataframe):
    df = zero_dataframe
    df_ams = am.score(df, ['z_score'])
    assert df_ams['z_score'].iloc[0] == 16.675431


#################
# LOGLIKELIHOOD #
#################

@pytest.mark.log_likelihood
def test_log_likelihood(fixed_dataframe):

    df = fixed_dataframe
    df_ams = am.score(df, ['log_likelihood'])
    assert df_ams['log_likelihood'][0] == 65.016595


@pytest.mark.log_likelihood
@pytest.mark.invalid
def test_log_likelihood_invalid(invalid_dataframe):
    df = invalid_dataframe
    with pytest.raises(ValueError):
        am.score(df, ['log_likelihood'])


@pytest.mark.log_likelihood
@pytest.mark.zero
def test_log_likelihood_zero(zero_dataframe):
    df = zero_dataframe
    df_ams = am.score(df, ['log_likelihood'], freq=True)
    assert df_ams['log_likelihood'].iloc[0] == 4087.276827


#############
# SIMPLE LL #
#############

@pytest.mark.simple_ll
def test_simple_ll(fixed_dataframe):

    df = fixed_dataframe
    df_ams = am.score(df, ['simple_ll'])
    assert df_ams['simple_ll'][0] == 28.051702


@pytest.mark.simple_ll
@pytest.mark.zero
def test_simple_ll_zero(zero_dataframe):
    df = zero_dataframe
    df_ams = am.score(df, ['simple_ll'], freq=True)
    assert df_ams['simple_ll'].iloc[0] == 264.915789


#############################
# hypergeometric likelihood #
#############################
# Not available via score due to numerical instability

@pytest.mark.choose
@pytest.mark.hypergeometric_likelihood
def test_hypergeometric_likelihood(fixed_dataframe):
    df = fq.expected_frequencies(fixed_dataframe, observed=True)
    df_ams = am.hypergeometric_likelihood(df)
    assert round(df_ams[0], 20) == 5.776904e-14


@pytest.mark.choose
@pytest.mark.hypergeometric_likelihood
def test_hypergeometric_likelihood_brown_overflow(brown_dataframe):
    df = fq.expected_frequencies(brown_dataframe, observed=True).head(10)
    df['hypergeometric_likelihood'] = am.hypergeometric_likelihood(df)
    assert df['hypergeometric_likelihood'].isnull().any()


@pytest.mark.choose
@pytest.mark.hypergeometric_likelihood
@pytest.mark.zero
def test_hypergeometric_likelihood_zero(zero_dataframe):
    df = fq.expected_frequencies(zero_dataframe, observed=True)
    ams = am.hypergeometric_likelihood(df)
    assert isnan(ams[0])


#######################
# binomial likelihood #
#######################
# Not available via score due to numerical instability

@pytest.mark.choose
@pytest.mark.binomial_likelihood
def test_binomial_likelihood(fixed_dataframe):
    df = fq.expected_frequencies(fixed_dataframe, observed=True)
    df_ams = am.binomial_likelihood(df)
    assert round(df_ams[0], 14) == 7.006036e-08


@pytest.mark.choose
@pytest.mark.binomial_likelihood
def test_binomial_likelihood_brown(brown_dataframe):
    df = brown_dataframe
    df = df.join(fq.observed_frequencies(df), rsuffix='_')
    df = df.join(fq.expected_frequencies(df), rsuffix='_')
    df = df.head(100)
    df['binomial_likelihood'] = am.binomial_likelihood(df)
    assert round(df['binomial_likelihood'][0], 6) == 0.008101


@pytest.mark.choose
@pytest.mark.binomial_likelihood
def test_binomial_likelihood_brown_overflow(brown_dataframe):
    df = brown_dataframe
    df = df.join(fq.observed_frequencies(df), rsuffix='_')
    df = df.join(fq.expected_frequencies(df), rsuffix='_')
    df = df.head(1000)
    df['binomial_likelihood'] = am.binomial_likelihood(df)
    assert df['binomial_likelihood'].isnull().any()


@pytest.mark.choose
@pytest.mark.binomial_likelihood
@pytest.mark.zero
def test_binomial_likelihood_zero(zero_dataframe):
    df = fq.expected_frequencies(zero_dataframe, observed=True)
    ams = am.binomial_likelihood(df)
    assert isnan(ams.iloc[0])


#############
# Log Ratio #
#############

@pytest.mark.log_ratio
def test_log_ratio(fixed_dataframe):

    df = fixed_dataframe
    df_ams = am.score(df, ['log_ratio'], disc=.5, discounting='Hardie2014')
    assert df_ams['log_ratio'].iloc[0] == 7.491853


@pytest.mark.log_ratio
@pytest.mark.invalid
def test_log_ratio_invalid(invalid_dataframe):

    df = invalid_dataframe
    with pytest.raises(ValueError):
        am.score(df, ['log_ratio'])


@pytest.mark.log_ratio
@pytest.mark.zero
def test_log_ratio_zero(zero_dataframe):

    df = zero_dataframe
    df_ams = am.score(df, ['log_ratio'], disc=.5, discounting='Hardie2014')
    assert df_ams['log_ratio'].iloc[0] == 12.03645


##########################
# Conservative Log Ratio #
##########################

@pytest.mark.conservative_log_ratio
def test_conservative_log_ratio(fixed_dataframe):

    df = fixed_dataframe
    df_ams = am.score(df, ['log_ratio', 'conservative_log_ratio'], boundary='normal', disc=.5, alpha=.01)
    assert (abs(df_ams['log_ratio']) >= abs(df_ams['conservative_log_ratio'])).all()
    assert df_ams['conservative_log_ratio'].iloc[0] == 0.796936


@pytest.mark.conservative_log_ratio
def test_conservative_log_ratio_zero(zero_dataframe):

    df = zero_dataframe
    df_ams = am.score(df, ['log_ratio', 'conservative_log_ratio'])
    assert (abs(df_ams['log_ratio']) >= abs(df_ams['conservative_log_ratio'])).all()


@pytest.mark.conservative_log_ratio
def test_conservative_log_ratio_zero_poisson(zero_dataframe):

    df = zero_dataframe
    df_ams = am.score(df, ['log_ratio', 'conservative_log_ratio'], boundary='poisson')
    assert (abs(df_ams['log_ratio']) >= abs(df_ams['conservative_log_ratio'])).all()


@pytest.mark.conservative_log_ratio
def test_conservative_log_ratio_zero_poisson_sig(zero_dataframe_sig):

    df = zero_dataframe_sig
    df_ams = am.score(df, ['log_ratio', 'conservative_log_ratio'], boundary='poisson')
    assert (abs(df_ams['log_ratio']) >= abs(df_ams['conservative_log_ratio'])).all()


@pytest.mark.conservative_log_ratio
def test_conservative_log_ratio_one_sided(fixed_dataframe):

    df = fq.expected_frequencies(fixed_dataframe, observed=True)
    df_ams = am.score(df, ['conservative_log_ratio'], boundary='normal')
    df_am = am.conservative_log_ratio(df, one_sided=True, boundary='normal')
    df_am.name = 'clr_one_sided'
    df_ams = df_ams.join(df_am)
    assert (abs(df_ams['conservative_log_ratio']) <= abs(df_ams['clr_one_sided'])).all()


@pytest.mark.conservative_log_ratio
def test_conservative_log_ratio_boundaries(brown_dataframe):

    df = brown_dataframe
    df_ams = am.score(df, ['conservative_log_ratio'])
    df_am = am.score(df, ['conservative_log_ratio'], boundary="normal")['conservative_log_ratio']
    df_am.name = 'clr_normal'
    df_ams = df_ams.join(df_am)
    assert (df_ams['clr_normal'] == 0).sum() < (df_ams['conservative_log_ratio'] == 0).sum()


###################
# MIN_SENSITIVITY #
###################

@pytest.mark.min_sensitivity
def test_min_sensitivity(fixed_dataframe):

    df = fixed_dataframe
    df_ams = am.score(df, ['min_sensitivity'])
    assert df_ams['min_sensitivity'][0] == 1


###########
# LIDDELL #
###########

@pytest.mark.liddell
def test_liddell(fixed_dataframe):

    df = fixed_dataframe
    df_ams = am.score(df, ['liddell'])
    assert df_ams['liddell'].iloc[0] == 1


@pytest.mark.liddell
def test_liddell_zero(zero_dataframe):

    df = fq.expected_frequencies(zero_dataframe, observed=True)
    df_ams = am.score(df, ['liddell'])
    assert df_ams['liddell'].iloc[0] == 0.143858


########
# GOLD #
########

@pytest.mark.gold
def test_measures_ucs_gold(ucs_dataframe):

    df = ucs_dataframe
    df = df.join(am.score(df, freq=False))

    for ucs, assoc in [('am.Dice', 'dice'),
                       ('am.MS', 'min_sensitivity'),
                       ('am.t.score', 't_score'),
                       ('am.z.score', 'z_score'),
                       ('am.MI', 'mutual_information'),
                       ('am.log.likelihood', 'log_likelihood'),
                       ('am.local.MI', 'local_mutual_information'),
                       ('am.simple.ll', 'simple_ll')]:

        assert round(df[ucs], 6).equals(df[assoc])


@pytest.mark.gold
def test_measures_log_ratio_gold(log_ratio_dataframe):

    df = log_ratio_dataframe
    df = df.join(am.score(df, ['log_ratio', 'conservative_log_ratio'], boundary='normal',
                          discounting='Hardie2014', disc=.5, alpha=.01, freq=False))

    for r, assoc in [('lr', 'log_ratio'),
                     ('clr', 'conservative_log_ratio')]:

        assert round(df[r], 6).equals(df[assoc])


@pytest.mark.gold
def test_measures_lrc_gold(log_ratio_dataframe):

    # original implementation with normal approximation
    df = log_ratio_dataframe
    df = df.join(am.score(df, ['conservative_log_ratio'], boundary='normal', alpha=.05, freq=False))
    assert df['conservative_log_ratio'].equals(round(df['lrc.normal'], 6))

    # implementation with poisson approximation
    df = log_ratio_dataframe
    df = df.join(am.score(df, ['conservative_log_ratio'],
                          alpha=.05, boundary='poisson', freq=False))
    assert df['conservative_log_ratio'].equals(round(df['lrc'], 6))


#################
# SCORE WRAPPER #
#################
@pytest.mark.score
def test_score_notation(ucs_dataframe):

    # frequency signature notation in dataframe:
    df1 = am.score(ucs_dataframe)

    # frequency signature notation with int parameters:
    f1 = int(ucs_dataframe['f1'].iloc[0])
    N = int(ucs_dataframe['N'].iloc[0])
    df2 = am.score(ucs_dataframe[['f', 'f2']], f1=f1, N=N)

    # corpus frequency notation in dataframe:
    tmp = ucs_dataframe[['f', 'f1']].rename({'f': 'f1', 'f1': 'N1'}, axis=1)
    tmp['N2'] = ucs_dataframe['N'] - ucs_dataframe['f1']
    tmp['f2'] = ucs_dataframe['f2'] - ucs_dataframe['f']
    df3 = am.score(tmp)

    # corpus frequency notation with int parameters:
    N1 = int(tmp['N1'].iloc[0])
    N2 = int(tmp['N2'].iloc[0])
    df4 = am.score(tmp[['f1', 'f2']], N1=N1, N2=N2)

    assert df1.equals(df2)
    assert df2.equals(df3)
    assert df3.equals(df4)


@pytest.mark.score
def test_score_invalid(ucs_dataframe):

    with pytest.raises(ValueError):
        am.score(ucs_dataframe, f1=1, N1=1)

    with pytest.raises(ValueError):
        am.score(ucs_dataframe, N2=1)

    with pytest.raises(ValueError):
        am.score(ucs_dataframe, f1=1, N2=1)


def test_calculate_measures(zero_dataframe):
    df = zero_dataframe
    with pytest.deprecated_call():
        df_ams = am.calculate_measures(df, ['dice'])
    df_ams['dice'].iloc[0] == 0.16831229174945742



#!/usr/bin/env python3
#cython: language_level=3
#cython: boundscheck=False
#cython: wraparound=False

from scipy.special.cython_special cimport beta, gamma

cpdef double cy_beta(double a, double b):
    return beta(a, b)

cpdef double complex cy_gamma(double complex z):
    return gamma(z)



"""
observed (O11, O12, O21, O22) and expected (E11, E12, E21, E22) frequencies

"""

from pandas import DataFrame


def observed_frequencies(df, f1=None, N=None, N1=None, N2=None, marginals=False):
    """Return observed frequencies in contingency table notation
    (O11..O22). Raises a Value Error if columns are not reasonably
    named.

    Notation for marginals (Evert 2008):
    - rows: R1 = O11 + O12, R2 = O21 + O22
    - columns: C1 = O11 + O21, C2 = O12 + O22
    - size: N = O11 + O12 + O21 + O22 = R1 + R2 = C1 + C2

    Possible input formats:
    - frequency signature (cf. Evert 2008: Figure 8):
      f  = O11                    # co-occurrence freq. of token and node / freq. in corpus 1
      f1 =  R1 <int>              # number of tokens in W(node) / size of corpus 1
      f2 =  C1                    # marginal freq. of token / freq. in corpus 1 + 2
      N  =   N <int>              # size of corpus without nodes / size of corpus 1 + 2
    - corpus frequencies ("keyword friendly"):
      f1 = O11                    # number of occurrences in corpus 1
      f2 = O21                    # number of occurrences in corpus 2
      N1 =  R1 <int>              # size of corpus 1
      N2 =  R2 <int>              # size of corpus 2

    Integers can also be passed as scalar arguments.

    :param DataFrame df: DataFrame with reasonably-named frequency columns
    :param bool marginals: add marginals? (R1, R2, C1, C2, N)
    :return: df with same index and columns O11, O12, O21, O22
    :rtype: DataFrame

    """

    # integer parameters instead of columns?
    if all(v is None for v in [f1, N1, N2, N]):
        pass

    elif f1 is not None and N is not None:
        try:
            df = df[['f', 'f2']].copy()
            df['f1'] = f1
            df['N'] = N
        except KeyError:
            raise ValueError(
                'frequency signature notation: (f1, N) are given, but (f, f2) are not'
            )

    elif N1 is not None and N2 is not None:
        try:
            df = df[['f1', 'f2']].copy()
            df['N1'] = N1
            df['N2'] = N2
        except KeyError:
            raise ValueError(
                'corpus frequency notation: (N1, N2) are given, but (f1, f2) are not'
            )

    else:
        raise ValueError(
            'either (f1, N) OR (N1, N2) have to be given'
        )

    # already in contingency notation?
    if set(['O11', 'O12', 'O21', 'O22']).issubset(df.columns):
        O11 = df['O11']
        O12 = df['O12']
        O21 = df['O21']
        O22 = df['O22']

    # frequency signature:
    elif set(['f', 'f1', 'f2', 'N']).issubset(df.columns):
        O11 = df['f']
        O12 = df['f1'] - O11
        O21 = df['f2'] - O11
        O22 = df['N'] - O11 - O12 - O21

    # corpus frequencies:
    elif set(['f1', 'N1', 'f2', 'N2']).issubset(df.columns):
        O11 = df['f1']
        O12 = df['N1'] - O11
        O21 = df['f2']
        O22 = df['N2'] - O21

    else:
        raise ValueError(f'columns are not reasonably named: {str(df.columns)}')

    # construct dataframe
    obs = DataFrame(
        index=df.index,
        data={
            'O11': O11,
            'O12': O12,
            'O21': O21,
            'O22': O22
        }
    )

    # add marginals
    if marginals:
        obs['R1'] = obs['O11'] + obs['O12']
        obs['R2'] = obs['O21'] + obs['O22']
        obs['C1'] = obs['O11'] + obs['O21']
        obs['C2'] = obs['O12'] + obs['O22']
        obs['N'] = obs['R1'] + obs['R2']

    return obs


def expected_frequencies(df, observed=False):
    """Calculate expected frequencies for observed frequencies assuming
    independence.

    :param pandas.DataFrame df: df with reasonably named columns
    :param bool observed: also return observed frequencies (contingency notations, incl. marginals)
    :return: df with same index and columns E11, E12, E21, E22
    :rtype: pandas.DataFrame

    """

    # convert input if necessary
    obs = observed_frequencies(df, marginals=True)

    # expected frequencies
    E11 = (obs['R1'] * obs['C1']) / obs['N']
    E12 = (obs['R1'] * obs['C2']) / obs['N']
    E21 = (obs['R2'] * obs['C1']) / obs['N']
    E22 = (obs['R2'] * obs['C2']) / obs['N']

    # construct dataframe
    expected = DataFrame(
        index=df.index,
        data={
            'E11': E11,
            'E12': E12,
            'E21': E21,
            'E22': E22
        }
    )

    # return marginals?
    df = obs.join(expected) if observed else expected

    return df



import pytest

from association_measures import binomial as bi


@pytest.mark.binomial
def test_choose():

    assert bi.choose(9, 3) == 84
    assert bi.choose(5, 2) == 10
    assert bi.choose(2, 2) == 1
    assert bi.choose(0, 0) == 1
    assert bi.choose(-3, 5) == 0
    assert bi.choose(-2, -2) == 0
    assert bi.choose(10, -2) == 0
    assert bi.choose(1000, 500) == 2.7028824094543644e+299



import association_measures.frequencies as fq


def test_observed_frequencies(fixed_dataframe):

    df = fq.observed_frequencies(fixed_dataframe)

    assert df['O11'][0] == 10
    assert df['O12'][0] == 0
    assert df['O21'][0] == 0
    assert df['O22'][0] == 90


def test_observed_frequencies_marginals(fixed_dataframe):

    df = fq.observed_frequencies(fixed_dataframe, marginals=True)

    assert df['O11'][0] == 10
    assert df['O12'][0] == 0
    assert df['O21'][0] == 0
    assert df['O22'][0] == 90
    assert df['N'][0] == 100
    assert df['C2'][0] == 90
    assert df['C1'][0] == 10
    assert df['R2'][0] == 90
    assert df['R1'][0] == 10


def test_expected_frequencies(fixed_dataframe):

    df = fq.expected_frequencies(fixed_dataframe)

    assert df['E11'].iloc[0] == 1.0


def test_expected_frequencies_observed(fixed_dataframe):

    df = fq.expected_frequencies(fixed_dataframe, observed=True)

    assert df['O11'].iloc[0] == 10
    assert df['E11'].iloc[0] == 1.0


def test_ucs(ucs_dataframe):

    df = ucs_dataframe

    # get observed frequencies
    df['O11'] = df['f']
    df['O21'] = df['f2'] - df['O11']
    df['O12'] = df['f1'] - df['O11']
    df['O22'] = df['N'] - df['f1'] - df['O21']

    # check observed frequencies
    obs = fq.observed_frequencies(df)
    assert obs['O11'].equals(df['O11'])
    assert obs['O12'].equals(df['O12'])
    assert obs['O21'].equals(df['O21'])
    assert obs['O22'].equals(df['O22'])

    # check marginals
    R1 = df['O11'] + df['O12']
    R2 = df['O21'] + df['O22']
    C1 = df['O11'] + df['O21']
    C2 = df['O12'] + df['O22']
    assert (R1 + R2).equals(df['N'])
    assert (C1 + C2).equals(df['N'])

    # get expected frequencies
    df['E11'] = R1 * C1 / df['N']
    df['E12'] = R1 * C2 / df['N']
    df['E21'] = R2 * C1 / df['N']
    df['E22'] = R2 * C2 / df['N']

    # check expected frequencies
    exp = fq.expected_frequencies(df)
    assert exp['E11'].equals(df['E11'])
    assert exp['E12'].equals(df['E12'])
    assert exp['E21'].equals(df['E21'])
    assert exp['E22'].equals(df['E22'])



#!/usr/bin/env python3

import os
import sys
from setuptools import find_packages, setup, Command, Extension

# Package meta-data.
NAME = 'association-measures'
DESCRIPTION = 'Statistical association measures for Python pandas'
URL = 'https://github.com/fau-klue/pandas-association-measures'
EMAIL = 'philipp.heinrich@fau.de'
AUTHOR = 'Philipp Heinrich & Markus Opolka'

REQUIRES_PYTHON = '>=3.9'
REQUIRED = [
    'wheel',
    'pandas>=2.2.2,<3.0',
    'numpy>=2.0,<3.0',
    'scipy>=1.13.0,<2.0',
    # 'rbo>=0.1.3,<0.2'
]

here = os.path.abspath(os.path.dirname(__file__))

# description
with open(os.path.join(here, 'README.md'), encoding="utf-8") as f:
    long_description = f.read()

# version
version = {}
with open(os.path.join(here, 'association_measures', 'version.py'), encoding="utf-8") as f:
    exec(f.read(), version)


# Import Cython if available
try:
    from Cython.Build import cythonize
    CYTHON_INSTALLED = True
    extensions = [Extension('association_measures.binomial', ['association_measures/binomial.pyx'])]
except ImportError:
    cythonize = lambda x, *args, **kwargs: x  # dummy func
    CYTHON_INSTALLED = False
    extensions = [Extension('association_measures.binomial', ['association_measures/binomial.c'])]


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            os.rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(version['__version__']))
        os.system('git push --tags')

        sys.exit()


setup(
    name=NAME,
    version=version['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=["tests", "test_*"]),
    ext_modules=cythonize(extensions),
    install_requires=REQUIRED,
    include_package_data=True,
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Cython',
    ],
)



from itertools import product

from numpy import exp, linspace, log
from pandas import DataFrame

from .measures import score


def expand_grid(dictionary):
    """Create a grid of all value combinations of all keys of the dictionary

    """

    return DataFrame([row for row in product(*dictionary.values())],
                     columns=dictionary.keys())


def log_seq(to=10e6, length=200, exact=50):
    """Create a logarithimcally scaled sequence

    """

    if length <= exact:
        raise ValueError()

    length = length - exact

    return list(range(exact + 1)) + [int(exp(s)) for s in sorted([x for x in linspace(log(exact), log(to), length)])]


def log_grid(N1=10e6, N2=10e6, length1=200, length2=200, exact1=50, exact2=50):
    """Create a logarithmically-scaled grid

    """
    return expand_grid({
        'f1': log_seq(N1, length1, exact=exact1),
        'f2': log_seq(N2, length2, exact=exact2)
    }).drop_duplicates().reset_index(drop=True)


def topography(N1=10e6, N2=10e6, length=200, length1=None, length2=None, exact=50, exact1=None, exact2=None):
    """Create logarithmically scaled grid and calculcate scores

    """

    exact1 = exact if exact1 is None else exact1
    exact2 = exact if exact2 is None else exact2
    length1 = length if length1 is None else length1
    length2 = length if length2 is None else length2

    # support
    g = log_grid(N1=N1, N2=N2, length1=length1, length2=length2, exact1=exact1, exact2=exact2)

    # add scores
    scores = score(g, N1=N1, N2=N2)
    # .. add alternative for CLR
    scores['clr_normal'] = score(
        g, N1=N1, N2=N2, boundary='normal', measures=['conservative_log_ratio']
    )['conservative_log_ratio']
    # .. add alternative for log-ratio
    scores['log_ratio_hardie'] = score(
        g, N1=N1, N2=N2, discounting='Hardie2014', measures=['log_ratio']
    )['log_ratio']

    return scores



#cython: language_level=3
"""
This file shows how the to use a BitGenerator to create a distribution.
"""
import numpy as np
cimport numpy as np
cimport cython
from cpython.pycapsule cimport PyCapsule_IsValid, PyCapsule_GetPointer
from libc.stdint cimport uint16_t, uint64_t
from numpy.random cimport bitgen_t
from numpy.random import PCG64
from numpy.random.c_distributions cimport (
      random_standard_uniform_fill, random_standard_uniform_fill_f)


@cython.boundscheck(False)
@cython.wraparound(False)
def uniforms(Py_ssize_t n):
    """
    Create an array of `n` uniformly distributed doubles.
    A 'real' distribution would want to process the values into
    some non-uniform distribution
    """
    cdef Py_ssize_t i
    cdef bitgen_t *rng
    cdef const char *capsule_name = "BitGenerator"
    cdef double[::1] random_values

    x = PCG64()
    capsule = x.capsule
    # Optional check that the capsule if from a BitGenerator
    if not PyCapsule_IsValid(capsule, capsule_name):
        raise ValueError("Invalid pointer to anon_func_state")
    # Cast the pointer
    rng = <bitgen_t *> PyCapsule_GetPointer(capsule, capsule_name)
    random_values = np.empty(n, dtype='float64')
    with x.lock, nogil:
        for i in range(n):
            # Call the function
            random_values[i] = rng.next_double(rng.state)
    randoms = np.asarray(random_values)

    return randoms

# cython example 2
@cython.boundscheck(False)
@cython.wraparound(False)
def uint10_uniforms(Py_ssize_t n):
    """Uniform 10 bit integers stored as 16-bit unsigned integers"""
    cdef Py_ssize_t i
    cdef bitgen_t *rng
    cdef const char *capsule_name = "BitGenerator"
    cdef uint16_t[::1] random_values
    cdef int bits_remaining
    cdef int width = 10
    cdef uint64_t buff, mask = 0x3FF

    x = PCG64()
    capsule = x.capsule
    if not PyCapsule_IsValid(capsule, capsule_name):
        raise ValueError("Invalid pointer to anon_func_state")
    rng = <bitgen_t *> PyCapsule_GetPointer(capsule, capsule_name)
    random_values = np.empty(n, dtype='uint16')
    # Best practice is to release GIL and acquire the lock
    bits_remaining = 0
    with x.lock, nogil:
        for i in range(n):
            if bits_remaining < width:
                buff = rng.next_uint64(rng.state)
            random_values[i] = buff & mask
            buff >>= width

    randoms = np.asarray(random_values)
    return randoms

# cython example 3
def uniforms_ex(bit_generator, Py_ssize_t n, dtype=np.float64):
    """
    Create an array of `n` uniformly distributed doubles via a "fill" function.

    A 'real' distribution would want to process the values into
    some non-uniform distribution

    Parameters
    ----------
    bit_generator: BitGenerator instance
    n: int
        Output vector length
    dtype: {str, dtype}, optional
        Desired dtype, either 'd' (or 'float64') or 'f' (or 'float32'). The
        default dtype value is 'd'
    """
    cdef Py_ssize_t i
    cdef bitgen_t *rng
    cdef const char *capsule_name = "BitGenerator"
    cdef np.ndarray randoms

    capsule = bit_generator.capsule
    # Optional check that the capsule if from a BitGenerator
    if not PyCapsule_IsValid(capsule, capsule_name):
        raise ValueError("Invalid pointer to anon_func_state")
    # Cast the pointer
    rng = <bitgen_t *> PyCapsule_GetPointer(capsule, capsule_name)

    _dtype = np.dtype(dtype)
    randoms = np.empty(n, dtype=_dtype)
    if _dtype == np.float32:
        with bit_generator.lock:
            random_standard_uniform_fill_f(rng, n, <float*>np.PyArray_DATA(randoms))
    elif _dtype == np.float64:
        with bit_generator.lock:
            random_standard_uniform_fill(rng, n, <double*>np.PyArray_DATA(randoms))
    else:
        raise TypeError('Unsupported dtype %r for random' % _dtype)
    return randoms

